import tkinter as tk


class SoundExplorer(object):
    def __init__(self, sound):
        self.sound = sound
        
    def show(self):
        sound = self.sound
        
        window = tk.Tk()
        window.wm_title(sound.filename.split('/')[-1] if hasattr(sound, 'filename') else 'Your sound!')
        
        width = 750
        height = 400
        
        canvas_frame = tk.Frame(window, width=width, height=height, 
                                highlightbackground='black', borderwidth=1, 
                                relief='sunken', padx=20)
        play_frame = tk.Frame(window, width=width, padx=100, pady=2)
        samples_frame = tk.Frame(window, width=width, pady=10)
        zoom_frame = tk.Frame(window, width=width)
        
        canvas = tk.Canvas(canvas_frame, highlightthickness=4, relief='ridge', 
                           highlightbackground='light gray',
                           width=width, height=height - 50,
                           bg='black')
        
        play_frame.grid(row=0, column=0)
        samples_frame.grid(row=1, column=0)
        canvas_frame.grid(row=2, column=0)
        zoom_frame.grid(row=3, column=0, pady=10)
        
        self.draw_lines(canvas, width, height, 0, len(sound.getSamples()))
        canvas.pack(pady=4)
        
        # This is a convoluted way of allowing two functions to reference each other.
        zoom_func_holder = [0]
        zoomed = False
        
        def zoom_out():
            nonlocal selected_region, zoomed
            
            self.draw_lines(canvas, width, height, 0, len(sound.getSamples()))
            zoom_button.config(text='Zoom In', state='disabled', command=zoom_func_holder[0])
            
            selected_region = [0, len(sound.samples)]
            
            text.config(state='normal')
            text.delete('1.0', tk.END)
            text.insert(tk.END, 'Samples [{} : {}]'.format(0, len(sound.getSamples())))
            text.tag_add('center', '1.0', 'end')
            text.config(state='disabled')
            
            zoomed = False
        
        selected_region = [0, len(sound.samples)]
        
        # Play button
        play_button = tk.Button(play_frame, text='Play Entire Sound', 
                  command=lambda: sound.play(),
                  width=16,
                  height=3,
                  relief='solid')
        play_button.grid(row=0, column=0, padx=8, pady=8)
        
        # Display sample index + sample value
        text = tk.Text(samples_frame, height=1, width=30, pady=6,
                       relief='solid')
        text.tag_configure('center', justify='center')
        text.insert(tk.END, 'Samples [{} : {}]'.format(0, len(sound.getSamples())))
        text.tag_add('center', '1.0', 'end')
        text.config(state='disabled')
        text.grid(row=2, column=0)
        
        start_x = 0
        selection_active = False
        
        start_sample = 0
        end_sample = 0
        
        # Play selected samples button        
        play_selection_button = tk.Button(play_frame, text='Play Selection',
                                          relief='solid', state='disabled',
                                          width=16,
                                          height=3,
                                          command=lambda: 
                                              sound.play(start_sample, end_sample))
        play_selection_button.grid(row=0, column=1)
        
        highlight_rect = canvas.create_rectangle(-1, -1, -1, -1, outline='cyan', fill='')
        def clear_selection():
            nonlocal highlight_rect, start_sample, end_sample
            
            play_selection_button.config(state='disabled')
            zoom_button.config(text='Zoom Out' if zoomed else 'Zoom In', 
                                   state='normal' if zoomed else 'disabled', 
                                   command=zoom_out)
            
            canvas.delete(highlight_rect)
            highlight_rect = canvas.create_rectangle(-1, -1, -1, -1, outline='cyan', fill='')
            
            clear_selection_button.config(state='disabled')
            text.config(state='normal')
            text.delete('1.0', tk.END)
            text.insert(tk.END, 'Samples [{} : {}]'.format(*selected_region))
            text.tag_add('center', '1.0', 'end')
            text.config(state='disabled')
        
        clear_selection_button = tk.Button(play_frame, text='Clear Selection',
                                           relief='solid', state='disabled',
                                           width=16,
                                           height=3,
                                           command=clear_selection)
        clear_selection_button.grid(row=1, column=0)
        
        def start_drag(event):
            nonlocal start_x, highlight_rect, selected_region, start_sample
            
            canvas.delete(highlight_rect)
            highlight_rect = canvas.create_rectangle(-1, -1, -1, -1, outline='cyan', fill='')
            begin, end = min(selected_region), max(selected_region)

            begin = int(begin + canvas.canvasx(event.x) * 
                           ((end-begin) / width))
            start_sample = begin
            
            start_x = canvas.canvasx(event.x)
            
            canvas.coords(highlight_rect, -1, -1, -1, -1)
        
        def mid_drag(event):
            nonlocal start_x, highlight_rect
            
            cur_x = canvas.canvasx(event.x)
            canvas.coords(highlight_rect, start_x, -10000, cur_x, 10000)
        
        def zoom():
            nonlocal selection_active, start_sample, end_sample, selected_region, zoomed
            
            self.draw_lines(canvas, width, height, 
                            start_sample, 
                            end_sample)
            
            selected_region = [start_sample, end_sample]
            
            zoom_button.config(text='Zoom Out', state='normal', command=zoom_out)
            play_selection_button.config(state='disabled')
            
            selection_active = False
            canvas.tag_raise(highlight_rect)
            
            zoomed = True
            clear_selection()
            
        zoom_func_holder[0] = zoom
        
        def end_drag(event):
            nonlocal selection_active, selected_region, start_sample, end_sample, zoomed
            
            begin, end = min(selected_region), max(selected_region)

            end = int(begin + canvas.canvasx(event.x) * 
                            ((end-begin) / width))
            
            # Just a click, not a drag.
            if end == start_sample:
                clear_selection()
                return
            
            end_sample = end
            start_sample, end_sample = min(start_sample, end_sample), max(start_sample, end_sample)
            
            sample_length = max(selected_region) - min(selected_region)
            if sample_length < 2: return
            
            text.config(state='normal')
            text.delete('1.0', tk.END)
            text.insert(tk.END, 'Samples [{} : {}]'.format(start_sample, end_sample))
            text.tag_add('center', '1.0', 'end')
            text.config(state='disabled')
            
            selection_active = True
            play_selection_button.config(state='normal')
            clear_selection_button.config(state='normal')
            zoom_button.config(state='normal', text='Zoom In', command=zoom)
            
        canvas.bind('<Button-1>', start_drag)
        canvas.bind('<B1-Motion>', mid_drag)
        canvas.bind('<ButtonRelease-1>', end_drag)
        
        zoom_button = tk.Button(
                zoom_frame, text='Zoom In', command=zoom,
                state='disabled',
                relief='solid',
                padx=16, pady=16)
        zoom_button.grid(row=0, column=0)
        
        exit_window = False
        def close_window():
            nonlocal exit_window
            exit_window = True
        
        window.protocol('WM_DELETE_WINDOW', close_window)
        while not exit_window:
            window.update()
            
        window.destroy()
        
    
    def draw_lines(self, canvas, width, height, begin, end):
        canvas.delete('all')
        length_factor = (end - begin) / width
        
        lines = []
        center = height // 2
        for index, s in enumerate(self.sound.getSamples()[begin:end]):
            val = center - (s.getValue() / (height / 2))
            i = index // length_factor
            
            if index > 0:
                lines.append([*prev, i, val])
            
            prev = i, val
            
        for line in lines:
            canvas.create_line(*line, width=1, fill='white')
            
        canvas.create_line(0, height // 2, 
                       width, height // 2, 
                       width=1, fill='cyan')
